from django.contrib.auth.mixins import (
    LoginRequiredMixin, PermissionRequiredMixin)
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.urls import reverse_lazy
from django.views.generic import FormView, UpdateView

from bursary.forms import (BursaryRefereeAddForm, BursaryUpdateForm,
                           BursaryMassUpdateForm)
from bursary.models import Bursary, BursaryReferee
from exports.views import CSVExportView


class BursaryAdminMixin(PermissionRequiredMixin):
    permission_required = 'bursary.add_bursaryreferee'


class BursaryRequestExport(BursaryAdminMixin, CSVExportView):
    model = Bursary
    filename = "bursaries.csv"
    ordering = ('user__username',)
    columns = [
        'user.username', 'user.email', 'user.first_name', 'user.last_name',
        'user.attendee.arrival', 'user.attendee.departure',
        'user.attendee.final_dates', 'user.attendee.reconfirm',
        'request_food', 'request_accommodation', 'request_travel',
        'reason_contribution', 'reason_plans', 'reason_diversity', 'need',
        'travel_bursary', 'travel_from',
        'accommodation_status', 'accommodation_accept_before',
        'food_status', 'food_accept_before',
        'travel_status', 'travel_accept_before',
        'reimbursed_amount',
    ]


class BursaryRefereeExport(BursaryAdminMixin, CSVExportView):
    model = BursaryReferee
    filename = "bursary_referees.csv"
    ordering = ('bursary__user__username', 'referee__username')
    columns = [
        'bursary.user.username', 'referee.username', 'final', 'contrib_score',
        'outreach_score', 'notes',
    ]


class BursaryRefereeAdd(BursaryAdminMixin, FormView):
    form_class = BursaryRefereeAddForm
    template_name = 'wafer/base_form.html'
    success_url = reverse_lazy('admin:bursary_bursaryreferee_changelist')

    def form_valid(self, form):
        new_referees = form.cleaned_data['to_add']
        referee_users = form.cleaned_data['referees']
        bursary_requests = form.cleaned_data['bursaries']
        for requestor, referees in new_referees.items():
            for referee in referees:
                BursaryReferee.objects.get_or_create(
                    bursary=bursary_requests[requestor],
                    referee=referee_users[referee],
                )
        return super(BursaryRefereeAdd, self).form_valid(form)


class BursaryMassUpdate(BursaryAdminMixin, FormView):
    form_class = BursaryMassUpdateForm
    template_name = 'wafer/base_form.html'
    success_url = reverse_lazy('admin:bursary_bursary_changelist')

    def form_valid(self, form):
        bursaries = form.cleaned_data['bursaries']
        notifies = form.cleaned_data['notifies']

        for bursary in bursaries:
            bursary.save()
            if bursary in notifies:
                bursary.notify_status(self.request)

        return super(BursaryMassUpdate, self).form_valid(form)


class BursaryUpdate(LoginRequiredMixin, SuccessMessageMixin, UpdateView):
    model = Bursary
    form_class = BursaryUpdateForm
    template_name = 'bursary/update.html'
    success_message = 'Bursary request updated successfully'

    def get_object(self):
        try:
            bursary = Bursary.objects.get(user=self.request.user)
        except Bursary.DoesNotExist:
            raise Http404('No bursary request submitted')
        else:
            if not bursary.request_any:
                raise Http404('Bursary request was withdrawn')
            if not bursary.can_update():
                raise Http404('Bursary cannot be updated')
            return bursary

    def get_success_url(self):
        return reverse_lazy('wafer_user_profile',
                            args=(self.request.user.username,))



import csv

from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import FieldDoesNotExist, ValidationError

from crispy_forms.helper import FormHelper
from crispy_forms.layout import ButtonHolder, Div, Fieldset, Layout, Submit

from bursary.models import Bursary


class BursaryRefereeAddForm(forms.Form):
    csv = forms.CharField(
        label='CSV input',
        help_text='Format: attendee_login,referee1_login,referee2_login',
        widget=forms.Textarea(attrs={'rows': 3}),
        required=True,
    )

    def __init__(self, *args, **kwargs):
        super(BursaryRefereeAddForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper(self)
        self.helper.include_media = False
        self.helper.add_input(Submit('submit', 'Create Referees'))

    def clean(self):
        cleaned_data = super(BursaryRefereeAddForm, self).clean()

        r = csv.reader(cleaned_data['csv'].splitlines())
        lines = {line[0]: set(line[1:]) for line in r}

        usernames = set(lines)
        referee_usernames = set().union(*lines.values())

        bursaries = {}
        missing_users = []

        referees = {}
        missing_referees = []

        for username in usernames:
            try:
                bursary = Bursary.objects.get(user__username=username)
            except Bursary.DoesNotExist:
                missing_users.append(username)
            else:
                bursaries[username] = bursary

        for username in referee_usernames:
            try:
                referee = User.objects.get(username=username)
            except User.DoesNotExist:
                missing_referees.append(username)
            else:
                if referee.has_perm('bursary.change_bursaryreferee'):
                    referees[username] = referee
                else:
                    missing_referees.append(username)

        errors = []

        if missing_users:
            errors.append(
                forms.ValidationError('no bursary request for users %s' %
                                      ', '.join(sorted(missing_users)))
            )

        if missing_referees:
            errors.append(
                forms.ValidationError('users %s are not referees' %
                                      ', '.join(sorted(missing_referees)))
            )

        if errors:
            raise forms.ValidationError(errors)

        cleaned_data['to_add'] = lines
        cleaned_data['bursaries'] = bursaries
        cleaned_data['referees'] = referees

        return cleaned_data


class BursaryMassUpdateForm(forms.Form):
    csv = forms.CharField(
        label='CSV input',
        help_text='add a header line',
        widget=forms.Textarea(attrs={'rows': 3}),
        required=True,
    )

    def __init__(self, *args, **kwargs):
        super(BursaryMassUpdateForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper(self)
        self.helper.include_media = False
        self.helper.add_input(Submit('submit', 'Update bursary requests'))

    def clean(self):
        cleaned_data = super(BursaryMassUpdateForm, self).clean()

        data = {}
        keys = set()
        r = csv.DictReader(cleaned_data['csv'].splitlines())
        for line in r:
            if 'user.username' not in line:
                raise forms.ValidationError(
                    'CSV data needs a `user.username` key'
                )
            data[line['user.username']] = line
            keys.update(line.keys())

        keys.remove('user.username')
        errors = []
        model_fields = {}
        for field in keys:
            try:
                model_fields[field] = Bursary._meta.get_field(field)
            except FieldDoesNotExist:
                errors.append(forms.ValidationError(
                    'Burary model has no %s field' % field
                ))
        if errors:
            raise forms.ValidationError(errors)

        bursaries = []
        notifies = []
        errors = []
        missing_users = []
        for username, updated_fields in data.items():
            try:
                bursary = Bursary.objects.get(user__username=username)
            except Bursary.DoesNotExist:
                missing_users.append(username)
                continue

            old_statuses = {
                'accommodation': bursary.accommodation_status,
                'food': bursary.food_status,
                'travel': bursary.travel_status,
            }

            for field, value in updated_fields.items():
                if field == 'user.username':
                    continue

                model_field = model_fields[field]
                if not value and model_field.null:
                    value = None

                try:
                    value = model_field.clean(value, bursary)
                except ValidationError as e:
                    errors.append(e)
                else:
                    setattr(bursary, field, value)

            try:
                bursary.full_clean()
            except ValidationError as e:
                errors.append(e)
                continue

            statuses = {
                'accommodation': bursary.accommodation_status,
                'food': bursary.food_status,
                'travel': bursary.travel_status,
            }
            if statuses != old_statuses:
                notifies.append(bursary)
            bursaries.append(bursary)

        if missing_users:
            errors.append(
                forms.ValidationError('no bursary request for users %s' %
                                      ', '.join(sorted(missing_users)))
            )

        if errors:
            raise forms.ValidationError(errors)

        cleaned_data['bursaries'] = bursaries
        cleaned_data['notifies'] = notifies

        return cleaned_data


class BursaryUpdateForm(forms.ModelForm):
    ALREADY_PROCESSED = 'Your reimbursement has already been processed'
    ONLY_DECREASE = ('You can only decrease the requested amount for your '
                     'travel bursrary')

    UPDATE_STATUS_CHOICES = (
        ('pending', 'Leave pending'),
        ('accepted', 'Accept'),
        ('canceled', 'Decline'),
    )

    class Meta:
        model = Bursary
        fields = ('accommodation_status', 'food_status', 'travel_status',
                  'travel_bursary')

    def __init__(self, *args, **kwargs):
        super(BursaryUpdateForm, self).__init__(*args, **kwargs)
        self.fields['accommodation_status'].choices = self.UPDATE_STATUS_CHOICES
        self.fields['accommodation_status'].label = 'Accommodation bursary'
        self.fields['food_status'].choices = self.UPDATE_STATUS_CHOICES
        self.fields['food_status'].label = 'Food bursary'
        self.fields['travel_status'].choices = self.UPDATE_STATUS_CHOICES
        self.fields['travel_status'].label = 'Travel bursary'
        self.fields['travel_bursary'].label = (
            'Final amount of my travel bursary request (in USD)'
        )
        self.fields['travel_bursary'].help_text = (
            'You can decrease the amount to that of your tickets once you '
            'know the final amount, so that more money is available for '
            'further rounds of bursaries. Food and accommodation are taken '
            'into account separately.'
        )

        acceptances = []
        for type_ in ('accommodation', 'food', 'travel'):
            field = '{}_status'.format(type_)
            if getattr(self.instance, field) == 'pending':
                acceptances.append(field)
            else:
                del self.fields[field]

        layout = []

        if acceptances:
            layout.append(
                Div(Fieldset('Accept or decline pending bursaries',
                             *acceptances),
                    css_class='alert alert-success')
            )

        if (self.instance.request_travel
                and self.instance.can_update('travel')
                and not self.instance.reimbursed_amount):
            layout.append(
                Fieldset('My travel bursary', 'travel_bursary')
            )
        else:
            del self.fields['travel_bursary']

        layout.append(ButtonHolder(Submit('submit', 'Update bursary status')))
        self.helper = FormHelper(self)
        self.helper.include_media = False
        self.helper.layout = Layout(*layout)

    def clean(self):
        cleaned_data = super(BursaryUpdateForm, self).clean()

        if (self.instance.request_travel and 'travel_bursary' in self.fields):

            if (self.instance.reimbursed_amount
                and (cleaned_data['travel_bursary']
                     != self.instance.travel_bursary)):
                raise forms.ValidationError(self.ALREADY_PROCESSED)

            if cleaned_data['travel_bursary'] > self.instance.travel_bursary:
                raise forms.ValidationError(self.ONLY_DECREASE)

        return cleaned_data



from django.conf.urls import url

from debconf.views import (
    DCScheduleArrived, DebConfScheduleView, IndexView, RobotsView
)

urlpatterns = [
    url(r'^attendees/admin/export/arrived/$', DCScheduleArrived.as_view()),
    url(r'^schedule/$', DebConfScheduleView.as_view(),
        name='wafer_full_schedule'),
    url(r'^robots.txt$', RobotsView.as_view()),
    url(r'^$', IndexView.as_view()),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-04 19:33
from __future__ import unicode_literals

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('reference_number', models.CharField(max_length=128, unique=True)),
                ('status', models.CharField(choices=[('new', 'Invoiced'), ('pending', 'Payment pending'), ('paid', 'Payment received'), ('canceled', 'Invoice canceled')], default='new', max_length=128)),
                ('date', models.DateField()),
                ('invoiced_entity', models.CharField(blank=True, max_length=128)),
                ('billing_address', models.TextField()),
                ('compound', models.BooleanField(default=False)),
                ('transaction_id', models.CharField(blank=True, max_length=128)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('recipient', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='invoices', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='InvoiceLine',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('line_order', models.IntegerField()),
                ('reference', models.CharField(max_length=32)),
                ('description', models.CharField(max_length=1024)),
                ('unit_price', models.DecimalField(decimal_places=2, max_digits=8)),
                ('quantity', models.IntegerField(validators=[django.core.validators.MinValueValidator(1)])),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='lines', to='invoices.Invoice')),
            ],
            options={
                'ordering': ('invoice', 'line_order'),
            },
        ),
        migrations.AlterUniqueTogether(
            name='invoiceline',
            unique_together=set([('invoice', 'line_order')]),
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import validators
from django.core.exceptions import ValidationError
from django.forms import Form, fields, widgets
from django.urls import reverse

from crispy_forms.helper import FormHelper
from crispy_forms.layout import ButtonHolder, HTML, Layout, Submit

from invoices.prices import MEAL_PRICES
from register.dates import meal_choices
from register.fields import MealSelectionField


class SearchForm(Form):
    q = fields.CharField(label='Find Users', required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_method = 'get'
        self.helper.include_media = False
        self.helper.add_input(Submit('search', 'Search'))


class CheckInForm(Form):
    t_shirt = fields.BooleanField(label='T-shirt', required=False)
    shoes = fields.BooleanField(label='Shoes', required=False)
    swag = fields.BooleanField(label='Swag Bag', required=False)
    nametag = fields.BooleanField(required=False)
    room_key = fields.BooleanField(required=False)
    key_card = fields.CharField(required=False)
    notes = fields.CharField(
        required=False,
        widget=widgets.Textarea(attrs={'rows': 3}),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.include_media = False
        if settings.DEBCONF_SHOE_SIZES:
            shoes = ('shoes',)
        else:
            shoes = ()
        self.helper.layout = Layout(
            't_shirt',
            *shoes,
            'swag',
            'nametag',
            'room_key',
            'key_card',
        )
        self.helper.layout.fields += [
            'notes',
            ButtonHolder(
                Submit('submit', 'Check In'),
                HTML('<a href="{}" class="btn btn-secondary">Cancel</a>'
                     .format(reverse('front_desk'))),
            ),
        ]


class CheckOutForm(Form):
    returned_key = fields.BooleanField(
        label='Returned Room Key', required=False)
    returned_card = fields.BooleanField(
        label='Returned Dorm Card', required=False)
    notes = fields.CharField(
        required=False,
        widget=widgets.Textarea(attrs={'rows': 3}),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.include_media = False
        self.helper.layout = Layout(
            'returned_key',
            'returned_card',
            'notes',
            ButtonHolder(
                Submit('submit', 'Check Out'),
                HTML('<a href="{}" class="btn btn-secondary">Cancel</a>'
                     .format(reverse('front_desk'))),
            ),
        )


class TShirtForm(Form):
    size = fields.ChoiceField(
        label='T-shirt',
        choices=settings.DEBCONF_T_SHIRT_SIZES,
        required=False,
    )

    def __init__(self, *args, username, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_action = reverse(
            'front_desk.change_shirt', kwargs={'username': username})
        self.helper.include_media = False
        self.helper.add_input(Submit('update', 'Update Shirt'))


class ShoesForm(Form):
    size = fields.ChoiceField(
        label='Shoes',
        choices=settings.DEBCONF_SHOE_SIZES,
        required=False,
    )

    def __init__(self, *args, username, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_action = reverse(
            'front_desk.change_shoes', kwargs={'username': username})
        self.helper.include_media = False
        self.helper.add_input(Submit('update', 'Update Shoes'))


class RegisterOnSiteForm(Form):
    name = fields.CharField()
    username = fields.CharField(validators=[
        validators.RegexValidator(
            r'^[\w.@+-]+$',
            'Enter a valid username. This value may contain only letters, '
            'numbers and @/./+/-/_ characters.',
            'invalid')
        ])
    email = fields.EmailField()
    departure = fields.DateTimeField()
    coc_ack = fields.BooleanField(label='Agrees to abide by CoC')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.include_media = False
        layout = self.helper.build_default_layout(self)
        layout.fields.append(
            ButtonHolder(
                Submit('register', 'Register'),
                HTML('<a href="{}" class="btn btn-secondary">Cancel</a>'
                     .format(reverse('front_desk'))),
            )
        )
        self.helper.add_layout(layout)

    def clean_username(self):
        username = self.cleaned_data['username']
        if get_user_model().objects.filter(username=username).exists():
            raise ValidationError('User with this username already exists')
        return username

    def clean_email(self):
        email = self.cleaned_data['email']
        if get_user_model().objects.filter(email=email).exists():
            raise ValidationError('User with this email already exists')
        return email


class CashInvoicePaymentForm(Form):
    receipt_number = fields.IntegerField()
    amount = fields.DecimalField(max_digits=8, decimal_places=2,
                                 validators=[validators.MinValueValidator(0)])

    def __init__(self, *args, instance, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance
        self.helper = FormHelper()
        self.helper.include_media = False
        layout = self.helper.build_default_layout(self)
        layout.fields.append(
            ButtonHolder(
                Submit('pay', 'Record Payment'),
                HTML(
                    '<a href="{}" class="btn btn-secondary">Cancel</a>'
                    .format(reverse(
                        'front_desk.check_in',
                        kwargs={'username': self.instance.recipient.username},
                    )))))
        self.helper.add_layout(layout)

    def clean_amount(self):
        amount = self.cleaned_data['amount']
        expected_amount = self.instance.total
        if amount != expected_amount:
            raise ValidationError('The invoice was for USD {}'
                                  .format(expected_amount))
        return amount


class FoodForm(Form):
    meals = fields.MultipleChoiceField(
        choices=meal_choices(),
        widget=widgets.CheckboxSelectMultiple,
        help_text="If you don't have a food bursary, meal prices are: "
                  "Breakfast {breakfast} USD, Lunch {lunch} USD, "
                  "Dinner {dinner} USD.".format(**MEAL_PRICES),
        required=False,
    )

    def __init__(self, *args, instance, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance
        self.helper = FormHelper()
        self.helper.include_media = False
        self.helper.layout = Layout(
            MealSelectionField('meals', id='meals'),
            HTML('<div class="alert alert-success">'
                 'Total: USD$<span id="total">0.00</span>. '
                 '<span style="display: none">'
                 'Difference: USD$<span id="delta">0.00</span>'
                 '</span></div>'),
            ButtonHolder(
                Submit('save', 'Save'),
                HTML(
                    '<a href="{}" class="btn btn-secondary">Cancel</a>'
                    .format(reverse(
                        'front_desk.check_in',
                        kwargs={'username': self.instance.user.username},
                    )))),
        )



from register.models.accommodation import Accomm, AccommNight
from register.models.attendee import Attendee
from register.models.childcare import ChildCare
from register.models.food import Food, Meal
from register.models.queue import Queue, QueueSlot
import register.speaker_stay

def user_is_registered(user):
    from register.views import STEPS
    last_step = len(STEPS) - 1
    return Attendee.objects.filter(
        user=user, completed_register_steps=last_step).exists()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-07-16 12:50
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('register', '0006_index_queue_monotonic_position'),
    ]

    operations = [
        migrations.AddField(
            model_name='attendee',
            name='completed_timestamp',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='attendee',
            name='created_timestamp',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='attendee',
            name='updated_timestamp',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



import datetime
from decimal import Decimal

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from invoices.models import Invoice, InvoiceLine
from register.dates import get_ranges_for_dates


INVOICE_PREFIX = settings.INVOICE_PREFIX
DEBCONF_NAME = settings.DEBCONF_NAME


FEE_INVOICE_INFO = {
    'pro': {
        'reference': INVOICE_PREFIX + 'REG-PRO',
        'description': '%s Professional registration fee' % DEBCONF_NAME,
        'unit_price': 200,
    },
    'corp': {
        'reference': INVOICE_PREFIX + 'REG-CORP',
        'description': '%s Corporate registration fee' % DEBCONF_NAME,
        'unit_price': 500,
    },
}

MEAL_PRICES = {
    'breakfast': None,
    'lunch': Decimal(5),
    'dinner': Decimal(5),
    'conference_dinner': Decimal(20),
}

ACCOMM_INVOICE_INFO = {
    'reference': INVOICE_PREFIX + 'ACCOMM',
    'description': '%s on-site accommodation' % DEBCONF_NAME,
    'unit_price': 5,
}


def invoice_user(user, force=False, save=False):
    from bursary.models import Bursary

    attendee = user.attendee

    try:
        bursary = user.bursary
    except Bursary.DoesNotExist:
        bursary = Bursary()

    lines = []
    fee_info = FEE_INVOICE_INFO.get(attendee.fee)
    if fee_info:
        fee_info['quantity'] = 1
        lines.append(InvoiceLine(**fee_info))

    try:
        accomm = attendee.accomm
    except ObjectDoesNotExist:
        accomm = None

    if accomm and not bursary.potential_bursary('accommodation'):
        for line in invoice_accomm(accomm):
            lines.append(InvoiceLine(**line))

    try:
        food = attendee.food
    except ObjectDoesNotExist:
        food = None

    if food and not bursary.potential_bursary('food'):
        for line in invoice_food(food):
            lines.append(InvoiceLine(**line))

    for paid_invoice in user.invoices.filter(status='paid', compound=False):
        lines.append(InvoiceLine(
            reference='INV#{}'.format(paid_invoice.reference_number),
            description='Previous Payment Received',
            unit_price=-paid_invoice.total,
            quantity=1,
        ))

    invoice = Invoice(
        recipient=user,
        status='new',
        date=timezone.now(),
        invoiced_entity=attendee.invoiced_entity,
        billing_address=attendee.billing_address
    )

    # Only save invoices if non empty
    if save and lines:
        invoice.save()

    total = 0
    for i, line in enumerate(lines):
        line.line_order = i
        total += line.total
        if save:
            line.invoice_id = invoice.id
            line.save()

    return {
        'invoice': invoice,
        'lines': lines,
        'total': total,
        'total_local': total * settings.DEBCONF_LOCAL_CURRENCY_RATE,
    }


def invoice_food(food):
    """Generate one invoice line per meal type per consecutive stay"""
    from register.models.food import Meal

    for meal, meal_label in Meal.MEALS.items():
        dates = [entry.date for entry in food.meals.filter(meal=meal)
                 if not entry.conference_dinner]
        if not dates:
            continue

        unit_price = MEAL_PRICES[meal]

        ranges = get_ranges_for_dates(dates)
        for first, last in ranges:
            n_meals = (last - first).days + 1

            if first != last:
                desc = "%s to %s" % (first, last)
            else:
                desc = str(first)

            full_desc = '%s %s (%s)' % (DEBCONF_NAME, meal_label, desc)
            yield {
                'reference': '%s%s' % (INVOICE_PREFIX, meal.upper()),
                'description': full_desc,
                'unit_price': unit_price,
                'quantity': n_meals,
            }

    if food.meals.filter(meal='dinner',
                         date=settings.DEBCONF_CONFERENCE_DINNER_DAY):
        yield {
            'reference': '%sCONFDINNER' % INVOICE_PREFIX,
            'description': '{} Conference Dinner ({})'.format(
                DEBCONF_NAME,
                settings.DEBCONF_CONFERENCE_DINNER_DAY.isoformat()),
            'unit_price': MEAL_PRICES['conference_dinner'],
            'quantity': 1,
        }


def invoice_accomm(accomm):
    """Generate one invoice line per consecutive stay"""
    stays = get_ranges_for_dates(
        night.date for night in accomm.nights.all()
    )

    for first_night, last_night in stays:
        last_morning = last_night + datetime.timedelta(days=1)
        num_nights = (last_morning - first_night).days
        desc = "evening of %s to morning of %s" % (first_night,
                                                   last_morning)

        line = ACCOMM_INVOICE_INFO.copy()
        line['description'] += ' ({})'.format(desc)
        line['quantity'] = num_nights
        yield line



from datetime import time
from django.conf import settings
from django.core.management.base import BaseCommand

from wafer.schedule.models import Venue, Day, Slot

import yaml


def parse_time(s):
    h, m = s.split(':')
    return time(int(h), int(m))


class Command(BaseCommand):
    help = 'Load Schedule days, venues, and slots from YAML files into the DB'

    def add_arguments(self, parser):
        parser.add_argument('YAMLFILE', help='Input file'),

    def handle(self, *args, **options):
        with open(options['YAMLFILE']) as f:
            data = yaml.safe_load(f)

        days = {}
        for datestr in data['days']:
            day, _ = Day.objects.get_or_create(date=datestr)
            days[datestr] = day
            print('Loaded day: %s' % day)

        for v in data['venues']:
            venue_name = v['name']
            venue, _ = Venue.objects.get_or_create(name=venue_name)
            venue.notes = v['notes']
            venue.order = v['order']
            venue_days = [day for d, day in days.items() if d in v['days']]
            venue.days = venue_days
            venue.save()
            print('Loaded venue: %s (%r)' % (venue, venue_days))

        for date, slots in data['slots'].items():
            day = days[date]
            for slotdata in slots:
                slot, _ = Slot.objects.get_or_create(day=day, name=slotdata['name'])
                slot.start_time = parse_time(slotdata['start'])
                slot.end_time = parse_time(slotdata['end'])
                slot.save()
                print('Loaded slot: %s' % slot)






from django.views.generic import FormView, DetailView
from django.views.generic.list import ListView
from django.views.generic.edit import UpdateView
from django.contrib import messages
from django.core.urlresolvers import reverse, reverse_lazy
from django.db.models import F, Q
from django.http import HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.shortcuts import redirect
from django import forms

from django.contrib.auth.decorators import permission_required

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit
from crispy_forms.bootstrap import FormActions

from wafer.users.views import EditOneselfMixin
from volunteers.models import Volunteer, Task
from volunteers.forms import (
    TasksFilterForm, VideoMassCreateTaskForm, VideoShirtForm)


class TasksView(ListView):
    model = Task
    template_name = 'wafer.volunteers/tasks.html'

    queryset = Task.objects.annotate_all()

    def get_context_data(self, **kwargs):
        context = super(TasksView, self).get_context_data(**kwargs)

        if self.request.GET:
            filter_form = TasksFilterForm(self.request.GET)
        else:
            filter_form = TasksFilterForm()

        if filter_form.is_valid():
            filter_data = filter_form.cleaned_data
        else:
            filter_data = filter_form.initial

        context['filter_form'] = filter_form

        tasks = context['object_list'].filter(
            start__lte=filter_data['end'],
            end__gte=filter_data['start'],
        )

        if 'needed' in filter_data['extra_filters']:
            tasks = tasks.filter(
                nbr_volunteers__lt=F('max_volunteers'),
            )

        context['tasks'] = tasks

        if (self.request.user.is_authenticated() and
                'preferred' in filter_data['extra_filters']):
            try:
                volunteer = Volunteer.objects.get(user=self.request.user)
                # TODO: add support for the preferred_tasks as well
                context['tasks'] = (
                    context['tasks'].filter(
                        Q(category__in=volunteer.preferred_categories.all())
                        | Q(template__in=volunteer.preferred_task_types.all())
                    )
                )
            except Volunteer.DoesNotExist:
                pass

        return context


class TaskView(DetailView):
    model = Task
    template_name = 'wafer.volunteers/task.html'

    queryset = Task.objects.annotate_all()

    def get_context_data(self, **kwargs):
        context = super(TaskView, self).get_context_data(**kwargs)
        # TODO Find a better way
        context['already_volunteer'] = (
            self.request.user.is_authenticated() and
            self.object.volunteers.filter(user=self.request.user).exists()
        )
        context['can_volunteer'] = (
            self.request.user.is_authenticated() and
            self.object.nbr_volunteers < self.object.max_volunteers
        )

        context['concurrent_tasks'] = self.object.concurrent_tasks()

        return context

    def post(self, request, *args, **kwargs):
        if not request.user.is_authenticated():
            return HttpResponseForbidden()
        self.object = self.get_object()
        volunteer, new = Volunteer.objects.get_or_create(user=request.user)
        concurrent_volunteers = [volunteer
                                 for task in self.object.concurrent_tasks()
                                 for volunteer in task.volunteers.all()]

        if self.object in volunteer.tasks.all():
            volunteer.tasks.remove(self.object)
            self.object.volunteers.remove(volunteer)
            messages.info(request, 'Removed')
        elif volunteer in concurrent_volunteers:
            # TODO Show where the user is going to be
            messages.error(request,
                'You are already volunteering in this time-slot. '
                'You can only be in one place at a time.')
        elif self.object.nbr_volunteers >= self.object.max_volunteers:
            messages.error(request, 'This task already has enough volunteers.')
        else:
            volunteer.tasks.add(self.object)
            self.object.volunteers.add(volunteer)
            messages.success(request, 'Thank you for volunteering!')

        return redirect('wafer_task', pk=self.object.pk)


class VolunteerView(EditOneselfMixin, DetailView):
    model = Volunteer
    slug_field = 'user__username'
    template_name = 'wafer.volunteers/volunteer.html'

    def get_object(self, *args, **kwargs):
        # TODO Find a better way
        if self.request.user.is_authenticated():
            Volunteer.objects.get_or_create(user=self.request.user)

        return super(VolunteerView, self).get_object(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(VolunteerView, self).get_context_data(**kwargs)
        context['profile'] = self.object.user.userprofile
        return context


class VolunteerUpdateForm(forms.ModelForm):
    class Meta:
        model = Volunteer
        fields = ['preferred_categories', 'preferred_task_types']
        widgets = {
            'preferred_categories': forms.CheckboxSelectMultiple(),
            'preferred_task_types': forms.CheckboxSelectMultiple(),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.helper = FormHelper()
        self.helper.layout = Layout(
            'preferred_categories',
            'preferred_task_types',
            FormActions(
                Submit('submit', 'Update preferences',
                       css_class='btn btn-success')
            ),
        )


class VolunteerUpdate(EditOneselfMixin, UpdateView):
    model = Volunteer
    slug_field = 'user__username'
    form_class = VolunteerUpdateForm
    template_name = 'wafer.volunteers/volunteer_update.html'

    def get_object(self, *args, **kwargs):
        # TODO Find a better way
        if self.request.user.is_authenticated():
            Volunteer.objects.get_or_create(user=self.request.user)

        return super(VolunteerUpdate, self).get_object(*args, **kwargs)

    def get_success_url(self):
        return reverse('wafer_volunteer',
                       kwargs={'slug': self.object.user.username})


class VideoMassScheduleView(FormView):
    @method_decorator(permission_required('volunteers.add_task'))
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    form_class = VideoMassCreateTaskForm
    template_name = 'wafer.volunteers/admin/video_mass_schedule.html'
    success_url = reverse_lazy('admin:volunteers_task_changelist')


class VideoShirtView(FormView):
    form_class = VideoShirtForm
    template_name = 'wafer.volunteers/admin/video_shirts.html'
    success_url = reverse_lazy('wafer_tasks')

    @method_decorator(permission_required('volunteers.add_task'))
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)



from collections import OrderedDict

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from invoices import prices

_pro_fee = prices.FEE_INVOICE_INFO['pro']['unit_price']
_corp_fee = prices.FEE_INVOICE_INFO['corp']['unit_price']


class Attendee(models.Model):
    FEES = OrderedDict((
        ('', 'Regular - Free'),
        ('pro', 'Professional - {} USD or {} {}'.format(
            _pro_fee, int(_pro_fee * settings.DEBCONF_LOCAL_CURRENCY_RATE),
            settings.DEBCONF_LOCAL_CURRENCY)),
        ('corp', 'Corporate - {} USD or {} {}'.format(
            _corp_fee, int(_corp_fee * settings.DEBCONF_LOCAL_CURRENCY_RATE),
            settings.DEBCONF_LOCAL_CURRENCY)),
    ))
    GENDERS = OrderedDict((
        ('', 'Decline to state'),
        ('m', 'Male'),
        ('f', 'Female'),
        ('n', 'Non-Binary'),
    ))

    user = models.OneToOneField(settings.AUTH_USER_MODEL,
                                related_name='attendee',
                                on_delete=models.PROTECT)

    # Contact information
    nametag_2 = models.CharField(max_length=50, blank=True)
    nametag_3 = models.CharField(max_length=50, blank=True)
    emergency_contact = models.TextField(blank=True)
    announce_me = models.BooleanField()
    register_announce = models.BooleanField()
    register_discuss = models.BooleanField()

    # Conference details
    coc_ack = models.BooleanField(default=False)
    fee = models.CharField(max_length=5, blank=True)
    arrival = models.DateTimeField(null=True, blank=True)
    departure = models.DateTimeField(null=True, blank=True)
    final_dates = models.BooleanField(default=False)
    reconfirm = models.BooleanField(default=False)

    # Personal information
    t_shirt_size = models.CharField(max_length=8, blank=True)
    shoe_size = models.CharField(max_length=8, blank=True)
    gender = models.CharField(max_length=1, blank=True)
    country = models.CharField(max_length=2, blank=True)
    languages = models.CharField(max_length=50, blank=True)
    pgp_fingerprints = models.TextField(blank=True)

    # Billing
    invoiced_entity = models.TextField(blank=True)
    billing_address = models.TextField(blank=True)

    # Misc
    notes = models.TextField(blank=True)
    completed_register_steps = models.IntegerField(default=0)
    created_timestamp = models.DateTimeField(auto_now_add=True)
    updated_timestamp = models.DateTimeField(auto_now=True)
    completed_timestamp = models.DateTimeField(null=True, blank=True)

    def __str__(self):
        return 'Attendee <{}>'.format(self.user.username)

    def billable(self):
        """Is this user billable? (or paid)"""
        from bursary.models import Bursary

        try:
            bursary = self.user.bursary
        except Bursary.DoesNotExist:
            bursary = Bursary()

        if self.fee:
            return True

        try:
            if (self.food.meals.exists()
                    and not bursary.potential_bursary('food')):
                return True
        except ObjectDoesNotExist:
            pass

        try:
            if (self.accomm.nights.exists()
                    and not bursary.potential_bursary('accommodation')):
                return True
        except ObjectDoesNotExist:
            pass

        return False

    billable.boolean = True

    def paid(self):
        from invoices.prices import invoice_user
        invoices = self.user.invoices
        if invoices.filter(status='new').exists():
            return False

        invoice = invoice_user(self.user)
        return invoice['total'] <= 0

    paid.boolean = True

    def registered_before_deadline(self):
        if not self.completed_timestamp:
            return False
        return (
            self.completed_timestamp.date() <=
            settings.DEBCONF_CONFIRMATION_DEADLINE)

    registered_before_deadline.boolean = True

    @property
    def new_invoices(self):
        return self.user.invoices.filter(status='new')

    @property
    def keysigning_id(self):
        from register.models.queue import QueueSlot
        try:
            keysigning_slot = QueueSlot.objects.get(
                queue__name='PGP Keysigning', attendee=self)
            return keysigning_slot.monotonic_position
        except QueueSlot.DoesNotExist:
            return None

    def save(self, *args, **kwargs):
        if self.arrival == '':
            self.arrival = None
        if self.departure == '':
            self.departure = None
        return super().save(*args, **kwargs)



from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import JsonResponse
from django.utils import timezone
from django.utils.crypto import constant_time_compare
from django.views.generic import TemplateView, View

from wafer.talks.models import Track
from wafer.schedule.views import ScheduleView
from wafer.schedule.models import Venue, ScheduleItem, Slot

from front_desk.models import CheckIn


class DCScheduleArrived(View):
    def arrived_users(self):
        queryset = CheckIn.objects.select_related(
            'attendee',
            'attendee__user',
            'attendee__user__userprofile',
        )
        for check_in in queryset.filter(attendee__announce_me=True):
            attendee = check_in.attendee
            user = attendee.user
            yield {
                'username': user.username,
                'name': user.userprofile.display_name(),
                'nick': attendee.nametag_3,
            }

    def get(self, request, *args, **kwargs):
        auth_header = request.META.get('HTTP_AUTHORIZATION', '')
        auth = auth_header.split(None, 1)
        if not constant_time_compare(auth, ['Bearer', settings.DCSCHEDULE_TOKEN]):
            raise PermissionDenied('Missing/Invalid Authorization token')
        return JsonResponse({
            'arrived': list(self.arrived_users())
        })


class RobotsView(TemplateView):
    template_name = 'debconf/robots.txt'
    content_type = 'text/plain; charset=UTF-8'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['SANDBOX'] = settings.SANDBOX
        return context


class DebConfScheduleView(ScheduleView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['tracks'] = Track.objects.all()
        return context


def get_current_slot():
    now = timezone.now()
    tz = timezone.get_default_timezone()

    for slot in Slot.objects.all():
        start = timezone.make_aware(slot.get_start_datetime(), tz)
        end = timezone.make_aware(slot.get_end_datetime(), tz)
        if start <= now and now < end:
            return slot


class IndexView(TemplateView):
    template_name = 'wafer/index.html'

    def get_context_data(self, *args, **kwargs):
        context_data = super().get_context_data(*args, **kwargs)

        venues = Venue.objects.filter(video=True)

        venue_blocks = [{'venue': venue} for venue in venues]

        current_slot = get_current_slot()
        if current_slot:
            events = ScheduleItem.objects.filter(slots=current_slot)
            for event in events:
                for blk in venue_blocks:
                    if event.venue == blk['venue']:
                        slots = list(event.slots.all())
                        blk['event'] = event
                        blk['start_time'] = slots[0].get_formatted_start_time()
                        blk['end_time'] = slots[-1].get_formatted_end_time()

        context_data['venue_blocks'] = venue_blocks

        return context_data



#!/usr/bin/env python
from setuptools import find_packages, setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='wafer-debconf',
    version='0.1.17',
    description='Wafer extensions used by DebConf',
    author='DebConf Team',
    author_email='debconf-team@lists.debian.org',
    url='https://salsa.debian.org/debconf-team/public/websites/wafer-debconf',
    packages=find_packages(),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Communications :: Conferencing',
    ],
)



import sys

from django.conf import settings
from django.contrib.auth.models import Group
from django.core.mail import EmailMultiAlternatives
from django.core.management.base import BaseCommand

from wafer.talks.models import Talk


FROM = 'content@debconf.org'

SUBJECT_A = '%(conference)s - talk accepted: %(title)s'
BODY_A = '''\
Dear speaker / BoF organizer,

We are glad to announce that your activity titled
%(title)s
was accepted for presentation at %(conference)s.

We look forward to seeing you in %(city)s.
We will soon start to put the schedule together, so make sure that your arrival
and departure dates are up to date in the system, so we can schedule your
activity on a suitable date.

In case your plans changed and you won't be attending anymore, please let us
know by replying to this email, as soon as possible. Also, please cancel your
conference registration in the conference website.

Best regards,
The %(conference)s Content Team
'''

SUBJECT_R = '%(conference)s - talk not accepted: %(title)s'
BODY_R = '''\
Dear submitter,

Thanks for your interest in %(conference)s.

We are sorry to say that your activity titled
%(title)s
was not accepted for presentation at the %(conference)s main schedule.

Note that %(conference)s will feature ad-hoc sessions where you can still
present your session. Those sessions are not covered by the video team, and so
not streamed or recorded. They are scheduled during the conference, so if you
are interested in presenting, come to %(conference)s, and and look out for an
email announcing the ad-hoc scheduling process.

We look forward to seeing you in %(city)s.

Best regards,
The %(conference)s Content Team
'''

TEMPLATES = {
    'A': {'subject': SUBJECT_A, 'body': BODY_A},
    'R': {'subject': SUBJECT_R, 'body': BODY_R},
}


OPTION_TO_STATUS = {
    'accepted': 'A',
    'not_accepted': 'R',
    'under_consideration': 'U',
    'provisionally_accepted': 'P',
}


class Command(BaseCommand):
    help = "Notify speakers about their talks."

    def add_arguments(self, parser):
        parser.add_argument('--yes', action='store_true',
                            help='Actually do something'),

        parser.add_argument('--accepted', action='store_true',
                            help='Notify accepted talks'),

        parser.add_argument('--not-accepted', action='store_true',
                            help='Notify rejected talks'),

        parser.add_argument('--provisionally-accepted', action='store_true',
                            help='Notify provisionally accepted talks'),

        parser.add_argument('--under-consideration', action='store_true',
                            help='Notify talks under consideration'),

        parser.add_argument('--subject', type=str, metavar='SUBJECT',
                            help='String to use as subject template for the emails'),

        parser.add_argument('--body', type=open, metavar='BODY',
                            help='File with template for the email body'),

        parser.add_argument('--bcc', type=str, nargs='+',
                            metavar='ADDR',
                            help='Add each ADDR to Bcc:'),

    def badger(self, talk, dry_run, subject_template, body_template, bcc):
        status = talk.get_status_display().lower().replace(' ', '_')
        kv, _ = talk.kv.get_or_create(
            group=self.content_group,
            key='notified_speaker_' + status,
            defaults={'value': None},
        )

        if kv.value:
            return

        to = [user.email for user in talk.authors.all()]

        subst = {
            'title': talk.title,
            'conference': settings.DEBCONF_NAME,
            'city': settings.DEBCONF_CITY,
        }

        subject = subject_template % subst
        body = body_template % subst

        if dry_run:
            print('I would badger speakers of: %s [status=%s]' % (talk.title,
                                                                  talk.status))
            return
        else:
            print('Badgering speakers of: %s' % talk.title)
        email_message = EmailMultiAlternatives(
            subject, body, from_email=FROM, to=to, bcc=bcc)
        email_message.send()

        kv.value = True
        kv.save()

    def handle(self, *args, **options):
        dry_run = not options['yes']
        self.content_group = Group.objects.get_by_natural_key('Talk Mentors')

        if dry_run:
            print('Not actually doing anything without --yes')

        status = None
        for key, value in OPTION_TO_STATUS.items():
            if options[key]:
                if status:
                    print('E: multiple talk statuses informed. Choose only one.')
                    sys.exit(1)
                status = value

        if not status:
            print("E: no talks selected; use --accepted, --not-accepted, etc")
            sys.exit(1)

        if status not in TEMPLATES:
            if not options['subject'] or not options['body']:
                print('E: no built-in template for status "%s"; please provide one explicitly with --subject and --body' % status)
                print('I: example subject template: %s' % TEMPLATES['A']['subject'])
                print('I: example body template:')
                print('----------------8<----------------8<----------------8<-----------------')
                print(TEMPLATES['A']['body'])
                print('----------------8<----------------8<----------------8<-----------------')
                sys.exit(1)

        if options['subject']:
            subject_template = options['subject']
        else:
            subject_template = TEMPLATES[status]['subject']

        if options['body']:
            body_template = options['body'].read()
        else:
            body_template = TEMPLATES[status]['body']

        bcc = options['bcc']

        for talk in Talk.objects.filter(status=status):
            self.badger(talk, dry_run, subject_template, body_template, bcc)



import yaml
from datetime import datetime
from dateutil.relativedelta import relativedelta
from django.utils import timezone
from django.core.management.base import BaseCommand, CommandError


from volunteers.models import TaskCategory, TaskTemplate, Task
from wafer.schedule.models import Venue


class Command(BaseCommand):
    help = 'Creates volunteer tasks from a YAML file'

    def add_arguments(self, parser):
        parser.add_argument('FILE', type=open)

    def handle(self, *args, **options):
        data = yaml.safe_load(options['FILE'])

        tasks = 0
        for item in data['tasks']:
            category, _ = TaskCategory.objects.get_or_create(name=item['category'])

            template, _ = TaskTemplate.objects.get_or_create(
                name=item['name']
            )
            if 'nbr_volunteers_min' in item:
                template.nbr_volunteers_min = item['nbr_volunteers_min']
            if 'nbr_volunteers_max' in item:
                template.nbr_volunteers_min = item['nbr_volunteers_max']
            if 'description' in item:
                template.description = item['description']
            if 'nbr_volunteers_max' in item or 'nbr_volunteers_max' in item or 'description' in item:
                template.save()

            if 'venues' in item:
                venues = []
                for v in item['venues']:
                    venue, _ = Venue.objects.get_or_create(name=v)
                    venues.append(venue)
            else:
                venues = [None]

            for day in item['days']:
                for hour in item['hours']:
                    startstr = hour['start']
                    endstr = hour['end']
                    start = timezone.make_aware(
                        datetime.strptime('%s %s' % (day, startstr), '%Y-%m-%d %H:%M')
                    )
                    end = timezone.make_aware(
                        datetime.strptime('%s %s' % (day, endstr), '%Y-%m-%d %H:%M')
                    )

                    for venue in venues:
                        task, _ = Task.objects.get_or_create(
                            template=template,
                            start=start,
                            end=end,
                            venue=venue,
                        )
                        print('Created task: %s' % task)
                        tasks = tasks + 1
        print('Tasks created: %d' % tasks)



from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied
from django.core.signing import BadSignature
from django.db import transaction
from django.forms.models import formset_factory
from django.http import Http404, HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.views.generic import DetailView, TemplateView

from paypal.standard.forms import PayPalPaymentsForm
from wafer.registration.views import redirect_profile

from invoices.forms import CombineForm, InvoiceForm, InvoiceFormset
from invoices.models import Invoice, InvoiceLine
from register.models import Attendee


class InvoiceDetailView(DetailView):
    model = Invoice
    slug_field = 'reference_number'
    slug_url_kwarg = 'reference_number'

    def get_object(self):
        invoice = super().get_object()

        if self.can_view_invoice(invoice):
            return invoice

        raise Http404()

    def can_view_invoice(self, invoice):
        user = self.request.user
        return (user.has_perm('front_desk.change_checkin')
                or invoice.recipient == user)

    def can_cancel_invoice(self, invoice):
        user = self.request.user
        return (user.has_perm('front_desk.change_invoice')
                or invoice.recipient == user)


class InvoiceDisplay(InvoiceDetailView):
    template_name = 'invoices/invoice_detail.html'

    def get_object(self):
        self.unsign()
        return super().get_object()

    def unsign(self):
        reference_number = self.kwargs['reference_number']
        self.using_signed_url = ':' in reference_number
        if self.using_signed_url:
            try:
                reference_number = Invoice.signer.unsign(reference_number)
            except BadSignature:
                raise Http404()
            self.kwargs['reference_number'] = reference_number

    def can_view_invoice(self, invoice):
        return self.using_signed_url or super().can_view_invoice(invoice)

    def get_context_data(self, *args, **kwargs):
        context_data = super().get_context_data(*args, **kwargs)

        paypal_data = self.object.paypal_data()
        paypal_data['notify_url'] = self.request.build_absolute_uri(
            reverse('paypal-ipn')
        )
        paypal_data['return_url'] = self.request.build_absolute_uri()

        context_data.update({
            'form': PayPalPaymentsForm(initial=paypal_data),
            'using_signed_url': self.using_signed_url,
            'DEBCONF_CITY': settings.DEBCONF_CITY,
            'DEBCONF_LOCAL_CURRENCY': settings.DEBCONF_LOCAL_CURRENCY,
            'DEBCONF_NAME': settings.DEBCONF_NAME,
        })

        return context_data


class InvoiceCancel(LoginRequiredMixin, InvoiceDetailView):

    @transaction.atomic
    def post(self, request, **kwargs):
        invoice = self.get_object()
        if not self.can_cancel_invoice(invoice):
            raise PermissionDenied()

        if invoice.status == 'new':
            invoice.status = 'canceled'
            invoice.save()
            messages.success(
                self.request, 'Invoice cancelled')
        else:
            messages.warning(
                self.request, 'Only unpaid invoices can be cancelled')

        return redirect_profile(self.request)


class FormFormsetView(TemplateView):
    def get_initial(self):
        return {}

    def get_initial_formset(self):
        return []

    def get_form_kwargs(self):
        kwargs = {
            'initial': self.get_initial(),
        }
        if self.request.method == 'POST':
            kwargs['data'] = self.request.POST
        return kwargs

    def get_form(self):
        return self.form_class(**self.get_form_kwargs())

    def get_formset_kwargs(self):
        kwargs = {
            'form_kwargs': self.get_formset_form_kwargs(),
        }
        if self.request.method == 'POST':
            kwargs['data'] = self.request.POST
        else:
            kwargs['initial'] = self.get_initial_formset()
        return kwargs

    def get_formset_form_kwargs(self):
        return {}

    def get_formset(self):
        Formset = formset_factory(
            self.formset_form_class, formset=self.formset_class, extra=2)
        return Formset(**self.get_formset_kwargs())

    def get_context_data(self, **kwargs):
        if 'form' not in kwargs:
            kwargs['form'] = self.get_form()
        if 'formset' not in kwargs:
            kwargs['formset'] = self.get_formset()
        return super().get_context_data(**kwargs)

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        formset = self.get_formset()
        if form.is_valid() and formset.is_valid():
            return self.forms_valid(form, formset)
        else:
            return self.forms_invalid(form, formset)

    def forms_valid(self, form, formset):
        return HttpResponseRedirect(self.get_success_url())

    def forms_invalid(self, form, formset):
        return self.render_to_response(
            self.get_context_data(form=form, formset=formset))


class InvoiceCombine(LoginRequiredMixin, FormFormsetView):
    template_name = 'invoices/combine.html'
    form_class = CombineForm
    formset_form_class = InvoiceForm
    formset_class = InvoiceFormset

    def get_initial(self):
        user = self.request.user
        initial = {
            'invoiced_entity': user.userprofile.display_name()
        }
        try:
            attendee = user.attendee
        except Attendee.DoesNotExist:
            pass
        else:
            initial['billing_address'] = attendee.billing_address
        return initial

    def get_initial_formset(self):
        return [{'url': url}
                for i, url in enumerate(self.request.GET.getlist('invoice'))]

    def get_formset_form_kwargs(self):
        return {
            'user': self.request.user,
        }

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)

    def forms_valid(self, form, formset):
        user = self.request.user

        invoice = Invoice(
            recipient=user,
            status='new',
            compound=True,
            date=timezone.now(),
            invoiced_entity=form.cleaned_data['invoiced_entity'],
            billing_address=form.cleaned_data['billing_address'],
        )
        invoice.save()

        for i, form in enumerate(formset.forms):
            if not hasattr(form, 'invoice'):
                continue
            child_invoice = form.invoice
            InvoiceLine(
                invoice_id=invoice.id,
                line_order=i,
                reference='INV#{}'.format(child_invoice.reference_number),
                description='DebConf attendance for {}'.format(
                    child_invoice.recipient.userprofile.display_name()
                ),
                unit_price=child_invoice.total,
                quantity=1,
            ).save()

        self.reference_number = invoice.reference_number
        return super().forms_valid(form, formset)

    def get_success_url(self):
        return reverse('invoices:display',
                       kwargs={'reference_number': self.reference_number})
